//
//  sctp helper functions
//   this code may be used under the terms of GPLv3
//   if you apply changes please sign the contributor license aggreement at https://www.elstel.org/mohican/ 
//   so that your changes can be included into the main trunk, (c) copyright by Elmar Stellnberger
//

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include "sockcommon.h"

int verbosity = 1; 

int setmicroseconds( struct timeval *tv, const char *mstxt, long jiffie ){
  long allms; char *endptr; int ret = 0;
  errno = 0; allms = strtol( mstxt, &endptr, 0 );
  if( errno != 0 || ( endptr != NULL && *endptr != '\000' ) ) return EINVAL;
  if( allms < 0 ) return E2BIG;  // time_t is known to be at least a long
  jiffie = 1000000 / jiffie; if( allms < jiffie ) { allms = jiffie; ret = ERANGE; }
  tv->tv_sec = (time_t) allms / 1000000; 
  tv->tv_usec =  (suseconds_t) allms % 1000000; 
  return ret;
}

#define NanoFraction 1000000000

int setnanoseconds( struct timespec *ts, const char *mstxt, long jiffie ){
  long long allns; char *endptr;
  errno = 0; allns = strtoll( mstxt, &endptr, 0 );
  if( errno != 0 || ( endptr != NULL && *endptr != '\000' ) ) return EINVAL;
  if( allns < 0 || allns / NanoFraction > MaxSigned(time_t) ) return E2BIG;
  ts->tv_sec = (time_t) allns / NanoFraction; 
  ts->tv_nsec =  (long) allns % NanoFraction; 
  return 0;
}

bool checkascii( const char *str ) {
  bool ok = true;
  while(*str) { ok = ok && *str>=' '; str++; }
  return ok;
}

int doclose(int fd) { int ret; while( ( ret = close(fd) ) && ret == EINVAL ); return ret; }
void dup2quitonerr( int oldfd, int newfd ) { int ret; do { ret = dup2(oldfd,newfd); } while( ret==EINVAL ); if(ret!=newfd) { perror("error by dup2 when daemonizing"); exit(2); } }

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


int parse_urlspec( char *url, struct UrlSpec *urlspec, bool force_ipv6 ) { 
  int i, j, k; long portnum; char *stopchar; int gai_errno; uint16_t port;
  uint16_t default_port = 80;
  i = strcspn(url,"/\\");
  if( i <= 0 || url[i-1] != ':' || url[i] != '/' || url[i+1] != '/' ) {
    fprintf(stderr,"expected protocol://path: %s\n",url);
    return 2;
  } else if ( i-1<4 || 5<i-1 || memcmp(url,"http",4) || ( i-1==5 && url[4]!='s' ) ) {
    fprintf(stderr,"unknown protocol: "); fwrite(url,i-1,1,stderr); fputc('\n',stderr);
    return 3;
  }
  urlspec->fullURL = url;
  switch(i-1) {
    case 4: urlspec->protocol = PR_HTTP; default_port = 80; break;
    case 5: urlspec->protocol = PR_HTTPS; default_port = 443; break;
    default: fprintf(stderr,"protocol (len=%i) not implemented\n",i); return 1;
  }
  k = i + 2 + strcspn(url+(i+2),"/");
  j = k - 1; while( url[j]!=':' && url[j]!='/' ) j--;
  if( url[j]==':' ) {
    errno = 0; portnum = strtol( url+j+1, &stopchar, 0 );
    if( errno!=0 || ( *stopchar && *stopchar != ':' && *stopchar != '\000' ) || portnum<=0 || portnum>=65535 ) {
      fprintf(stderr,"error: '"); fwrite( url+j+1, k-(j+1), 1, stderr); fprintf(stderr,"' is not a valid port number.\n\n");
      return 1;
    } else port = (uint16_t) portnum;
    url[j] = '\000';
  } else {
    port = default_port;
    j = k;
  }
  if( port == default_port ) urlspec->port_override = -1;
  else urlspec->port_override = port;
  urlspec->host = (const char*)malloc(j-i);  // i is one char before start: accounts for len
  memcpy( (char*)(urlspec->host), url+i+2, j-i-1);
  ((char*)urlspec->host)[j-i-2] = '\000';
  if( url[k]=='\000' || !strcmp(url+k,"/") ) {
    urlspec->path = "/";
    urlspec->target = NULL;
  } else {
    urlspec->path = url + k;
    urlspec->target = strrchr(url+k,'/'); if(urlspec->target) { urlspec->target++; if(*urlspec->target=='\000') urlspec->target=NULL; }
  }
  struct addrinfo template;
  memset( &template, 0, sizeof(template) ); 
  template.ai_flags = AI_V4MAPPED | AI_ALL;  // no IPv6 found: return IPv4-mapped IPv6 addresses  +AI_ALL return both IPv6 and IPv4-mapped IPv6 addresses
  template.ai_family = force_ipv6 ? AF_INET6 : AF_UNSPEC; template.ai_socktype = SOCK_STREAM; template.ai_protocol = 0;  //  that would be for tcp but we will not care about this value later on
  urlspec->addr = NULL; gai_errno = getaddrinfo( urlspec->host, NULL, &template, &urlspec->addr );
  if(gai_errno) { fprintf(stderr,"getaddrinfo(%s): %s.\n\n",urlspec->host, gai_strerror(gai_errno) ); return 2; }
  if(!urlspec->addr) { fprintf(stderr,"getaddrinfo(%s) returned no result.\n\n",urlspec->host); return 2; }
  struct sockaddr *adr = urlspec->addr->ai_addr;
  if(adr->sa_family==AF_INET) ((struct sockaddr_in*)adr)->sin_port = htons( port );
  else if(adr->sa_family==AF_INET6) ((struct sockaddr_in6*)adr)->sin6_port = htons( port );
  else { fprintf(stderr,"unsupported protocol (neither IPv4 nor IPv6)\n"); return 2; }
  return 0;
}

void print_sockaddr( FILE* out, struct sockaddr *sa ) {
  if( sa->sa_family == AF_INET ) {
    struct sockaddr_in* sa4 = (struct sockaddr_in*)sa;
    long inaddr = ntohl(sa4->sin_addr.s_addr);
    if( inaddr == INADDR_ANY ) fprintf(out,"*");
    else while(inaddr) { fprintf(out,"%i", (int) inaddr & 0xFF ); inaddr >>= 8; if(inaddr) fputc('.',out); }
    fprintf(out,":%i",ntohs(sa4->sin_port));
  } else if( sa->sa_family == AF_INET6 ) {
    struct sockaddr_in6 *sa6 = (struct sockaddr_in6*)sa;
    fprintf(out,"?:%i",ntohs(sa6->sin6_port));
  } else fprintf(out,"unknown/non-IP address family");
}

const char* protocol2str[3] = { NULL, "http", "https" };

void print_urlspec( FILE *out, struct UrlSpec *urlspec ) {
  fprintf(out,"protocol: %s\n",protocol2str[urlspec->protocol]);
  fprintf(out,"addr: "); print_sockaddr(out,urlspec->addr->ai_addr); fputc('\n',out);
  fprintf(out,"host: %s\n", urlspec->host);
  fprintf(out,"path: %s\n", urlspec->path);
  fprintf(out,"target: %s\n\n", urlspec->target ? urlspec->target : "NULL");
}


int socket_connect( int *s, struct addrinfo *target_host_ais, int flags ) {
  struct addrinfo *hst; int saved_errno = ESRCH; bool silent = ( flags >> 2 ) & 1;
  assert(!(flags&1));

  for( hst = target_host_ais; hst; hst = hst->ai_next ) {
    if( connect( *s, hst->ai_addr, hst->ai_addrlen ) == 0 ) break;
    saved_errno = errno; 
    if( !silent ) perror("error trying to connect tcp socket to address"); 
    errno = saved_errno;
  }

  if ( *s <= -2 ) saved_errno = EAGAIN;    // EXDEV ~ improper link
  // *s == -1: shield for a timeout signal handler which could have set s=-2 or any other value below zero
  if( !hst ) return saved_errno;

  return 0;
}


